/*
 * (C) Copyright 2007-2012
 * Allwinner Technology Co., Ltd. <www.allwinnertech.com>
 * Benn Huang <benn@allwinnertech.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/string.h>
#include <linux/device.h>
#include <linux/dma-mapping.h>
#include <linux/platform_device.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/amba/bus.h>
#include <linux/amba/clcd.h>
#include <linux/amba/pl061.h>
#include <linux/amba/mmci.h>
#include <linux/amba/pl022.h>
#include <linux/io.h>
#include <linux/gfp.h>
#include <linux/clockchips.h>
#include <linux/memblock.h>
#include <linux/bootmem.h>
#include <linux/export.h>
#include <linux/clkdev.h>

#include <asm/setup.h>
#include <asm/sizes.h>
#include <asm/mach-types.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/leds.h>
#include <asm/setup.h>
#include <asm/memory.h>
#include <asm/delay.h>
#include <asm/mach/map.h>
#include <asm/mach/arch.h>
#include <asm/mach/flash.h>
#include <asm/mach/irq.h>
#include <asm/mach/time.h>
#include <asm/hardware/arm_timer.h>
#include <asm/hardware/icst.h>
#include <asm/hardware/vic.h>

#include <plat/core.h>
#include <plat/memory.h>
#include <plat/platform.h>
#include <plat/system.h>
#include <plat/sys_config.h>

struct brom_header {
	u32 jump_instruction;	/* one intruction jumping to real code */
	char magic[8];		/* ="eGON.BRM",  not C-style string. */
	u32 length;		/* generated by PC */
	char Boot_vsn[4];	/* Boot version */
	char eGON_vsn[4];	/* eGON version */
	char platform[8];	/* platform information */
};

static inline const struct brom_header *_sunxi_brom(void)
{
	const struct brom_header *brom = (void*)SW_VA_BROM_BASE + 0x4000;
	if (strncmp(brom->magic, "eGON.BRM", 8) == 0)
		return brom;
	else
		return NULL;
}

int sunxi_pr_brom(void)
{
	const struct brom_header *brom = _sunxi_brom();
	if (brom) {
		pr_info("BROM Ver: %.4s %.4s %.8s\n",
			brom->Boot_vsn,
			brom->eGON_vsn,
			brom->platform);
		return 1;
	}
	pr_err("SUNXI BROM not found");
	return 0;
}

u32 sunxi_chip_id(void)
{
	const struct brom_header *brom = _sunxi_brom();
	if (brom) {
		char *endptr;
		u32 chip_id = simple_strtoul(brom->platform, &endptr, 10);
		if (*endptr == '\0')
			return chip_id;

		pr_err("BROM: failed to parse chip-id: \"%s\"\n",
		       brom->platform);
	}
	return 0;
}

int sunxi_pr_chip_id()
{
	u32 chip_id = sunxi_chip_id();
	const char *name;

	switch (chip_id) {
	case SUNXI_CHIP_ID_A10:
		name = "A10"; break;
	case SUNXI_CHIP_ID_A13:
		name = "A13"; break;
	default:
		name = "Unknown";
	}

	if (machine_is_sun4i())
		pr_info("chip-id: %s (AW%u revision %c)\n", name,
			chip_id, 'A' + sw_get_ic_ver());
	else
		pr_info("chip-id: %s (AW%u)\n", name, chip_id);

	return (chip_id != 0);
}

/*
 * Only reserve certain important memory blocks if there are actually
 * drivers which use them.
 */
static int reserved_mali_mem;
static unsigned long reserved_start;
static unsigned long reserved_max;

#ifdef CONFIG_SUNXI_MALI_RESERVED_MEM
void __init sunxi_mali_core_fixup(struct tag *tags, char **cmdline,
				  struct meminfo *mi)
{
	struct tag *t;
	u32 bank = 0;

	for (t = tags; t->hdr.size; t = tag_next(t)) {
		if (t->hdr.tag != ATAG_CMDLINE)
			continue;
		if (strstr(t->u.cmdline.cmdline, "sunxi_no_mali_mem_reserve"))
			return;
	}

	for (t = tags; t->hdr.size; t = tag_next(t)) {
		if (t->hdr.tag != ATAG_MEM)
			continue;
		if (bank) {
			mi->nr_banks++;
			mi->bank[bank].start = t->u.mem.start;
			mi->bank[bank].size = t->u.mem.size;
		} else { /* first bank */
			u32 size = t->u.mem.size / SZ_1M;
			mi->nr_banks = 1;
			mi->bank[0].start = t->u.mem.start;
			if (size < 512) {
				mi->bank[0].size = SZ_1M * size;

				pr_err("MALI: not enough memory in first bank to make reserve.\n");
			} else {
				mi->bank[0].size = SZ_1M * (512 - 64);
				reserved_mali_mem = 1;
				size -= 512;
				if (size) {
					bank++;
					mi->nr_banks++;
					mi->bank[1].start = t->u.mem.start + (512 * SZ_1M);
					mi->bank[1].size = SZ_1M * size;
				}

				pr_info("Memory cut off:\n");
				pr_reserve_info("MALI", t->u.mem.start + SZ_512M - SZ_64M,
						SZ_64M);
			}
		}
		bank++;
	}
}
#endif

/**
 * Machine Implementations
 *
 */

static struct map_desc sw_io_desc[] __initdata = {
	{ SW_VA_SRAM_BASE, __phys_to_pfn(SW_PA_SRAM_BASE),  (SZ_128K + SZ_64K), MT_MEMORY_ITCM  },
	{ SW_VA_IO_BASE,   __phys_to_pfn(SW_PA_IO_BASE),    (SZ_1M + SZ_2M),    MT_DEVICE       },
	{ SW_VA_BROM_BASE, __phys_to_pfn(SW_PA_BROM_BASE),  (SZ_64K),           MT_MEMORY_ITCM  },
};

void __init sw_core_map_io(void)
{
	iotable_init(sw_io_desc, ARRAY_SIZE(sw_io_desc));

	sunxi_pr_chip_id();
}

#ifdef CONFIG_FB_SUNXI_RESERVED_MEM
/* The FB block is used by:
 *
 * - the sun4i framebuffer driver, drivers/video/sun4i/disp.
 *
 * fb_start, fb_size are used in a vast number of other places but for
 * for platform-specific drivers, so we don't have to worry about them.
 */

unsigned long fb_start;
unsigned long fb_size = SZ_32M;
EXPORT_SYMBOL(fb_start);
EXPORT_SYMBOL(fb_size);

static int __init reserve_fb_param(char *s)
{
	unsigned long size;
	if (kstrtoul(s, 0, &size) == 0)
		fb_size = size * SZ_1M;
	return 0;
}
early_param("sunxi_fb_mem_reserve", reserve_fb_param);
#endif

#if defined CONFIG_SUN4I_G2D || defined CONFIG_SUN4I_G2D_MODULE
/* The G2D block is used by:
 *
 * - the G2D engine, drivers/char/sun4i_g2d
 */

unsigned long g2d_start;
unsigned long g2d_size = SZ_1M * 16;
EXPORT_SYMBOL(g2d_start);
EXPORT_SYMBOL(g2d_size);

static int __init reserve_g2d_param(char *s)
{
	unsigned long size;
	if (kstrtoul(s, 0, &size) == 0)
		g2d_size = size * SZ_1M;
	return 0;
}
early_param("sunxi_g2d_mem_reserve", reserve_g2d_param);
#endif

#if defined CONFIG_VIDEO_DECODER_SUN4I || \
	defined CONFIG_VIDEO_DECODER_SUN4I_MODULE || \
	defined CONFIG_VIDEO_DECODER_SUN5I || \
	defined CONFIG_VIDEO_DECODER_SUN5I_MODULE
/* The VE block is used by:
 *
 * - the Cedar video engine, drivers/media/video/sun4i
 */

#define RESERVE_VE_MEM 1

unsigned long ve_start;
unsigned long ve_size = (SZ_64M + SZ_16M);
EXPORT_SYMBOL(ve_start);
EXPORT_SYMBOL(ve_size);

static int __init reserve_ve_param(char *s)
{
	unsigned long size;
	if (kstrtoul(s, 0, &size) == 0)
		ve_size = size * SZ_1M;
	return 0;
}
early_param("sunxi_ve_mem_reserve", reserve_ve_param);
#endif

static void reserve_sys(void)
{
	memblock_reserve(SYS_CONFIG_MEMBASE, SYS_CONFIG_MEMSIZE);
	pr_reserve_info("SYS ", SYS_CONFIG_MEMBASE, SYS_CONFIG_MEMSIZE);
}

#if defined RESERVE_VE_MEM || defined CONFIG_SUN4I_G2D || \
	defined CONFIG_SUN4I_G2D_MODULE || defined CONFIG_FB_SUNXI_RESERVED_MEM
static void reserve_mem(unsigned long *start, unsigned long *size,
			const char *desc)
{
	if (*size == 0) {
		*start = 0;
		return;
	}

	if ((reserved_start + *size) > reserved_max) {
		pr_warn("Not enough memory to reserve memory for %s\n", desc);
		*start = 0;
		*size = 0;
		return;
	}
	*start = reserved_start;
	memblock_reserve(*start, *size);
	pr_reserve_info(desc, *start, *size);
	reserved_start += *size;
}
#endif

static void __init sw_core_reserve(void)
{
	pr_info("Memory Reserved:\n");
	reserve_sys();
	/* 0 - 64M is used by reserve_sys */
	reserved_start = meminfo.bank[0].start + SZ_64M;
	reserved_max   = meminfo.bank[0].start + meminfo.bank[0].size;
#ifdef CONFIG_FB_SUNXI_RESERVED_MEM
	if (reserved_mali_mem) {
		/* The stupid mali blob expects the fb at a fixed address :( */
		fb_start = meminfo.bank[0].start + SZ_512M - SZ_64M - SZ_32M;
		if (fb_start < reserved_max)
			reserved_max = fb_start;
	}
#endif
#ifdef RESERVE_VE_MEM
	reserve_mem(&ve_start, &ve_size, "VE  ");
#endif
#if defined CONFIG_SUN4I_G2D || defined CONFIG_SUN4I_G2D_MODULE
	reserve_mem(&g2d_start, &g2d_size, "G2D ");
#endif
#ifdef CONFIG_FB_SUNXI_RESERVED_MEM
	if (reserved_mali_mem) {
		/* Give the mali blob the fb at its expected address */
		reserved_start = fb_start;
		reserved_max   = meminfo.bank[0].start + meminfo.bank[0].size;
	}
	reserve_mem(&fb_start, &fb_size, "LCD ");
#endif
}

void sw_irq_ack(struct irq_data *irqd)
{
	unsigned int irq = irqd->irq;

	if (irq < 32){
		writel(readl(SW_INT_ENABLE_REG0) & ~(1<<irq), SW_INT_ENABLE_REG0);
		writel(readl(SW_INT_MASK_REG0) | (1 << irq), SW_INT_MASK_REG0);
		writel(readl(SW_INT_IRQ_PENDING_REG0) | (1<<irq), SW_INT_IRQ_PENDING_REG0);
	} else if(irq < 64){
		irq -= 32;
		writel(readl(SW_INT_ENABLE_REG1) & ~(1<<irq), SW_INT_ENABLE_REG1);
		writel(readl(SW_INT_MASK_REG1) | (1 << irq), SW_INT_MASK_REG1);
		writel(readl(SW_INT_IRQ_PENDING_REG1) | (1<<irq), SW_INT_IRQ_PENDING_REG1);
	} else if(irq < 96){
		irq -= 64;
		writel(readl(SW_INT_ENABLE_REG2) & ~(1<<irq), SW_INT_ENABLE_REG2);
		writel(readl(SW_INT_MASK_REG2) | (1 << irq), SW_INT_MASK_REG2);
		writel(readl(SW_INT_IRQ_PENDING_REG2) | (1<<irq), SW_INT_IRQ_PENDING_REG2);
	}
}

/* Mask an IRQ line, which means disabling the IRQ line */
static void sw_irq_mask(struct irq_data *irqd)
{
	unsigned int irq = irqd->irq;

	if(irq < 32){
		writel(readl(SW_INT_ENABLE_REG0) & ~(1<<irq), SW_INT_ENABLE_REG0);
		writel(readl(SW_INT_MASK_REG0) | (1 << irq), SW_INT_MASK_REG0);
	} else if(irq < 64){
		irq -= 32;
		writel(readl(SW_INT_ENABLE_REG1) & ~(1<<irq), SW_INT_ENABLE_REG1);
		writel(readl(SW_INT_MASK_REG1) | (1 << irq), SW_INT_MASK_REG1);
	} else if(irq < 96){
		irq -= 64;
		writel(readl(SW_INT_ENABLE_REG2) & ~(1<<irq), SW_INT_ENABLE_REG2);
		writel(readl(SW_INT_MASK_REG2) | (1 << irq), SW_INT_MASK_REG2);
	}
}

static void sw_irq_unmask(struct irq_data *irqd)
{
	unsigned int irq = irqd->irq;

	if(irq < 32){
		writel(readl(SW_INT_ENABLE_REG0) | (1<<irq), SW_INT_ENABLE_REG0);
		writel(readl(SW_INT_MASK_REG0) & ~(1 << irq), SW_INT_MASK_REG0);
		if(irq == SW_INT_IRQNO_ENMI) /* must clear pending bit when enabled */
			writel((1 << SW_INT_IRQNO_ENMI), SW_INT_IRQ_PENDING_REG0);
	} else if(irq < 64){
		irq -= 32;
		writel(readl(SW_INT_ENABLE_REG1) | (1<<irq), SW_INT_ENABLE_REG1);
		writel(readl(SW_INT_MASK_REG1) & ~(1 << irq), SW_INT_MASK_REG1);
	} else if(irq < 96){
		irq -= 64;
		writel(readl(SW_INT_ENABLE_REG2) | (1<<irq), SW_INT_ENABLE_REG2);
		writel(readl(SW_INT_MASK_REG2) & ~(1 << irq), SW_INT_MASK_REG2);
	}
}

static struct irq_chip sw_vic_chip = {
	.name       = "sw_vic",
	.irq_ack    = sw_irq_ack,
	.irq_mask   = sw_irq_mask,
	.irq_unmask = sw_irq_unmask,
};

void __init sw_core_init_irq(void)
{
	u32 i = 0;

	/* Disable & clear all interrupts */
	writel(0, SW_INT_ENABLE_REG0);
	writel(0, SW_INT_ENABLE_REG1);
	writel(0, SW_INT_ENABLE_REG2);

	writel(0xffffffff, SW_INT_MASK_REG0);
	writel(0xffffffff, SW_INT_MASK_REG1);
	writel(0xffffffff, SW_INT_MASK_REG2);

	writel(0xffffffff, SW_INT_IRQ_PENDING_REG0);
	writel(0xffffffff, SW_INT_IRQ_PENDING_REG1);
	writel(0xffffffff, SW_INT_IRQ_PENDING_REG2);
	writel(0xffffffff, SW_INT_FIQ_PENDING_REG0);
	writel(0xffffffff, SW_INT_FIQ_PENDING_REG1);
	writel(0xffffffff, SW_INT_FIQ_PENDING_REG2);

	/*enable protection mode*/
	writel(0x01, SW_INT_PROTECTION_REG);
	/*config the external interrupt source type*/
	writel(0x00, SW_INT_NMI_CTRL_REG);

	for (i = SW_INT_START; i < SW_INT_END; i++) {
		irq_set_chip(i, &sw_vic_chip);
		irq_set_handler(i, handle_level_irq);
		set_irq_flags(i, IRQF_VALID | IRQF_PROBE);
	}
}



/*
 * Global vars definitions
 *
 */
static void sun4i_restart(char mode, const char *cmd)
{
	/* use watch-dog to reset system */
	#define WATCH_DOG_CTRL_REG  (SW_VA_TIMERC_IO_BASE + 0x0094)
	*(volatile unsigned int *)WATCH_DOG_CTRL_REG = 0;
	__delay(100000);
	*(volatile unsigned int *)WATCH_DOG_CTRL_REG = 3;
	while(1);
}

static void timer_set_mode(enum clock_event_mode mode, struct clock_event_device *clk)
{
	volatile u32 ctrl;

	switch (mode) {
	case CLOCK_EVT_MODE_PERIODIC:
		pr_info("timer0: Periodic Mode\n");
		writel(TMR_INTER_VAL, SW_TIMER0_INTVAL_REG); /* interval (999+1) */
		ctrl = readl(SW_TIMER0_CTL_REG);
		ctrl &= ~(1<<7);    /* Continuous mode */
		ctrl |= 1;  /* Enable this timer */
		break;

	case CLOCK_EVT_MODE_ONESHOT:
		pr_info("timer0: Oneshot Mode\n");
		ctrl = readl(SW_TIMER0_CTL_REG);
		ctrl |= (1<<7);     /* Single mode */
		break;
	case CLOCK_EVT_MODE_UNUSED:
	case CLOCK_EVT_MODE_SHUTDOWN:
	default:
		ctrl = readl(SW_TIMER0_CTL_REG);
		ctrl &= ~(1<<0);    /* Disable timer0 */
		break;
	}

	writel(ctrl, SW_TIMER0_CTL_REG);
}

/* Useless when periodic mode */
static int timer_set_next_event(unsigned long evt, struct clock_event_device *unused)
{
	volatile u32 ctrl;

	/* clear any pending before continue */
	ctrl = readl(SW_TIMER0_CTL_REG);
	writel(evt, SW_TIMER0_CNTVAL_REG);
	ctrl |= (1<<1);
	writel(ctrl, SW_TIMER0_CTL_REG);
	writel(ctrl | 0x1, SW_TIMER0_CTL_REG);

	return 0;
}

static struct clock_event_device timer0_clockevent = {
	.name = "timer0",
	.shift = 32,
	.rating = 100,
	.features = CLOCK_EVT_FEAT_PERIODIC | CLOCK_EVT_FEAT_ONESHOT,
	.set_mode = timer_set_mode,
	.set_next_event = timer_set_next_event,
};


static irqreturn_t sw_timer_interrupt(int irq, void *dev_id)
{
	struct clock_event_device *evt = (struct clock_event_device *)dev_id;

	writel(0x1, SW_TIMER_INT_STA_REG);

	/*
 	 * timer_set_next_event will be called only in ONESHOT mode
 	 */
	evt->event_handler(evt);

	return IRQ_HANDLED;
}

static struct irqaction sw_timer_irq = {
	.name = "timer0",
	.flags = IRQF_DISABLED | IRQF_TIMER | IRQF_IRQPOLL,
	.handler = sw_timer_interrupt,
	.dev_id = &timer0_clockevent,
	.irq = SW_INT_IRQNO_TIMER0,
};


static void __init sw_timer_init(void)
{
	int ret;
	volatile u32  val = 0;

	writel(TMR_INTER_VAL, SW_TIMER0_INTVAL_REG);
	/* set clock sourch to HOSC, 16 pre-division */
	val = readl(SW_TIMER0_CTL_REG);
	val &= ~(0x07<<4);
	val &= ~(0x03<<2);
	val |= (4<<4) | (1<<2);
	writel(val, SW_TIMER0_CTL_REG);
	/* set mode to auto reload */
	val = readl(SW_TIMER0_CTL_REG);
	val |= (1<<1);
	writel(val, SW_TIMER0_CTL_REG);

	ret = setup_irq(SW_INT_IRQNO_TIMER0, &sw_timer_irq);
	if (ret) {
		pr_warning("failed to setup irq %d\n", SW_INT_IRQNO_TIMER0);
	}

	/* Enable time0 interrupt */
	val = readl(SW_TIMER_INT_CTL_REG);
	val |= (1<<0);
	writel(val, SW_TIMER_INT_CTL_REG);

	timer0_clockevent.mult = div_sc(SYS_TIMER_CLKSRC/SYS_TIMER_SCAL, NSEC_PER_SEC, timer0_clockevent.shift);
	timer0_clockevent.max_delta_ns = clockevent_delta2ns(0xff, &timer0_clockevent);
	timer0_clockevent.min_delta_ns = clockevent_delta2ns(0x1, &timer0_clockevent);
	timer0_clockevent.cpumask = cpumask_of(0);
	timer0_clockevent.irq = sw_timer_irq.irq;
	clockevents_register_device(&timer0_clockevent);
}

struct sys_timer sw_sys_timer = {
	.init = sw_timer_init,
};

extern void __init sw_pdev_init(void);
void __init sw_core_init(void)
{
	sw_pdev_init();
}
enum sw_ic_ver sw_get_ic_ver(void)
{
	volatile u32 val = readl(SW_VA_TIMERC_IO_BASE + 0x13c);

	val = (val >> 6) & 0x3;

	if (machine_is_sun4i()) {
		switch (val) {
		case 0x00:
			return MAGIC_VER_A;
		case 0x03:
			return MAGIC_VER_B;
		default:
			return MAGIC_VER_C;
		}
	} else {
		switch (val) {
		case 0x03:
			return MAGIC_VER_B;
		default:
			return MAGIC_VER_A;
		}
	}
}
EXPORT_SYMBOL(sw_get_ic_ver);

MACHINE_START(SUN4I, "sun4i")
	.atag_offset	= 0x100,
	.timer          = &sw_sys_timer,
#ifdef CONFIG_SUNXI_MALI_RESERVED_MEM
	.fixup          = sunxi_mali_core_fixup,
#endif
	.map_io         = sw_core_map_io,
	.init_early     = NULL,
	.init_irq       = sw_core_init_irq,
	.init_machine   = sw_core_init,
	.reserve        = sw_core_reserve,
	.restart	= sun4i_restart,
MACHINE_END

MACHINE_START(SUN5I, "sun5i")
	.atag_offset	= 0x100,
	.timer          = &sw_sys_timer,
#ifdef CONFIG_SUNXI_MALI_RESERVED_MEM
	.fixup          = sunxi_mali_core_fixup,
#endif
	.map_io         = sw_core_map_io,
	.init_early     = NULL,
	.init_irq       = sw_core_init_irq,
	.init_machine   = sw_core_init,
	.reserve        = sw_core_reserve,
	.restart	= sun4i_restart,
MACHINE_END
